------------------------------------------------------------------------------
------------------------------------------------------------------------------
--                                                                          --
--                                                                          --
-- Copyright (c) 2008 Tobias Gubener <tobiflex@opencores.org>               -- 
--                                                                          --
-- This source file is free software: you can redistribute it and/or modify --
-- it under the terms of the GNU General Public License as published        --
-- by the Free Software Foundation, either version 3 of the License, or     --
-- (at your option) any later version.                                      --
--                                                                          --
-- This source file is distributed in the hope that it will be useful,      --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of           --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            --
-- GNU General Public License for more details.                             --
--                                                                          --
-- You should have received a copy of the GNU General Public License        --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.    --
--                                                                          --
------------------------------------------------------------------------------
------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.STD_LOGIC_UNSIGNED.all;

 
entity membooster is
   port ( 
    reset: in std_logic;		--H aktiv
	clk: in std_logic;
	e: in std_logic;
	cpuaddress: in std_logic_vector(23 downto 1);
	cpudata_out: out std_logic_vector(15 downto 0);
	cpudata_in: in std_logic_vector(15 downto 0);
	address_agnus: in std_logic_vector(20 downto 1);
	dma: in std_logic;
	dmawr: in std_logic;
	dmapri: in std_logic;
	ovl: in std_logic;
	l_as: in std_logic;
	l_lds: in std_logic;
	l_uds: in std_logic;
	lr_w: in std_logic;
	l_dtack: buffer std_logic;
	cpurd: buffer std_logic;
	cpuhwr: buffer std_logic;
	cpulwr: buffer std_logic;
	cpuok: buffer std_logic;
	extra_chip: in std_logic;
	extra_slow: in std_logic;
	aron: in std_logic;
	boot: in std_logic;
	datain: in std_logic_vector(15 downto 0);	 	--bus data in
	dataout: out std_logic_vector(15 downto 0);		--bus data out
	cdataout: out std_logic_vector(15 downto 0);		--bus data out
	rd: buffer std_logic;			   				--bus read
	hwr: buffer std_logic;							--bus high write
	lwr: buffer std_logic;							--bus low write
	n_ub: out std_logic;							--sram upper byte
	n_lb: out std_logic;   							--sram lower byte
	n_we: out std_logic;							--sram write enable
	n_oe: out std_logic;    						--sram output enable
	data: in std_logic_vector(15 downto 0);	  		--sram data
--	mxaddress: buffer std_logic_vector(20 downto 1);	--chipram address 	 
	ramaddress: out std_logic_vector(23 downto 1);	--sdram address 	 
	selreg: buffer std_logic;							--
	selciaa: buffer std_logic;							--
	selciab: buffer std_logic;							--
--	selboot: out std_logic;							--
	
	selide: buffer std_logic;
	ideirq: in std_logic;
	iderdy: in std_logic;
	idedata: in std_logic_vector(15 downto 0);	  		--ide data
	intide: out std_logic;	
	
	cpu_dma: out std_logic;	
	cpuenable:buffer std_logic
	
   );

end membooster;

architecture wire of membooster is

COMPONENT bootrom 
	PORT
	(
		address		: IN STD_LOGIC_VECTOR (10 DOWNTO 0);
		clock		: IN STD_LOGIC ;
		q			: OUT STD_LOGIC_VECTOR (15 DOWNTO 0)
	);
END COMPONENT;

signal enable: std_logic;
signal selchip: std_logic;
signal selslow: std_logic;
signal selfast: std_logic;
signal selkick: std_logic;
signal selboot: std_logic;
--signal selkickovl: std_logic;
signal selgayleID: std_logic;
signal selgayle: std_logic;
signal shiftgid: std_logic_vector(7 downto 0);
signal ecpu: std_logic;
--signal cpuenable: std_logic;

--Autoconfig
   SIGNAL auto_enable : std_logic;
   SIGNAL nibble      : std_logic_vector(3 downto 0);
   SIGNAL cfgout      : std_logic;
   SIGNAL selautoconfig      : std_logic;
   SIGNAL cfgin       : std_logic;
   
   SIGNAL int2ena      : std_logic;   
   SIGNAL irq_pena      : std_logic;   
   SIGNAL irq_flag      : std_logic;   
   SIGNAL ideirqd      : std_logic;   
   SIGNAL ideirqdd      : std_logic;   
   
   
signal bootdata: std_logic_vector(15 downto 0);

begin
------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------
-- bootrom
bootrom_inst: bootrom
		PORT MAP (
		clock => NOT clk,
		address => cpuaddress(11 downto 1),
		q => bootdata
	);

------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------
-- cpu bridge

cdataout <= cpudata_in WHEN cpuhwr='1' OR cpulwr='1' ELSE "0000000000000000";

-- generate cpurd,cpuhwr,cpulwr and enable
process(cpuok,l_as,l_dtack,l_uds,l_lds,e)
begin
	cpuenable <= '0';
	cpu_dma <= '0'; 
	cpurd <= '0';
	cpuhwr <= '0';
	cpulwr <= '0';	
	-- normal cpu cycle
	IF cpuok='1' AND l_as='0' AND l_dtack='1' AND (l_uds='0' OR l_lds='0') THEN
		cpuenable <= '1';
		IF selboot='0' AND selgayleID='0' AND selgayle='0' AND selide='0' AND auto_enable='0' THEN
			cpu_dma <= '1'; 
		END IF;
		IF lr_w='1' THEN	-- 16bit read cycle
			cpurd <= '1';
		ELSE				-- 16bit or 8bit write cycle
			IF l_uds='0' THEN
				cpuhwr <= '1';
			END IF;
			IF l_lds='0' THEN
				cpulwr <= '1';
			END IF;
		END IF;
	END IF;
end process; 

-- dtack control
process(clk)
begin
	IF clk'event AND clk = '1' THEN
		IF cpuenable='1' THEN
			cpudata_out <= datain;
			IF selboot='1' THEN
				cpudata_out <= bootdata;
			END IF;
			IF selgayleID='1' THEN
				cpudata_out(15) <= shiftgid(7);
			END IF;
			IF selgayle='1' THEN
--				cpudata_out <= idedata;
				cpudata_out(15 downto 8) <="00000000";
--				cpudata_out(15 downto 0) <="1111111111111111";
--	when (A==DA8) then
--		d15=irq_passing_enabled.fb # irq_synched.fb;
				IF cpuaddress(13 downto 12)="00"THEN	--DA8000
--					cpudata_out(15) <= irq_pena OR ideirqd;
--					cpudata_out(15) <= NOT irq_pena AND ideirqd;
					cpudata_out(15) <= ideirqd;
				END IF;
--	when (A==DA9) then
--		d14='0';
--	else when (A==DA9) then
--		d15=interrupt_flag.fb; "Access interrupt flaggie
				IF cpuaddress(13 downto 12)="01"THEN	--DA9000
					cpudata_out(15) <= irq_flag;
--					cpudata_out(14) <= '0';
				END IF;
--	else when (A==DAA) then
--		d15=jeejee.fb;
				IF cpuaddress(13 downto 12)="10"THEN	--DAA000
					cpudata_out(15) <= int2ena;
				END IF;
			ELSIF selide='1' THEN
				cpudata_out <= idedata;
			END IF;
--	else when (A==DE) then
--		d15=d15_id_output.fb; "Access hardware identification
			IF auto_enable='1' THEN
				cpudata_out(15 downto 12) <= NOT nibble;
			END IF;
		END IF;	
	END IF;
	
	IF clk'event AND clk = '1' THEN
		IF l_as='1' THEN
			l_dtack <= '1';
		ELSIF cpuenable='1' THEN
			l_dtack <= '0';
		END IF;
	END IF;
end process; 
------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------
--gayle
--gayle ID at $de1000 (write here => shiftreg init)
--gayle ID is $D0
-- read here => shift one bit
--IDE da0000-da3fff
--gayle address decode
process(clk)
begin
	IF reset='1' THEN
		irq_flag <= '0';
		irq_pena <= '0';
		int2ena <= '0';
	ELSIF clk'event AND clk = '1' THEN
		ideirqd <= ideirq OR irq_pena;   
		ideirqdd <= ideirqd;   
--	else when (irq_rose.fb) then
--		interrupt_flag := !irq_passing_enabled.fb;
		IF (ideirqd='1' AND ideirqdd='0') then --OR (ideirqd='0' AND ideirqdd='1') THEN
--			irq_flag <= NOT irq_pena;-- OR irq_flag;
			irq_flag <= int2ena;
--			irq_flag <= NOT int2ena;
--			irq_flag <= '1';--NOT int2ena;
		END IF;
--		IF cpuaddress(13 downto 12)="01" AND selgayle='1' AND cpuok='1' THEN
--			irq_flag <= '0';
--		END IF;
		IF selgayle='1' AND (cpuhwr='1' OR cpulwr='1')THEN
--	when ((its_me_lasted.fb==2) & (A==DA8) & (r_w.pin=='0')) then
--		irq_passing_enabled := d15.pin;	
			IF cpuaddress(13 downto 12)="00" THEN	--DA8000
				irq_pena <= cpudata_in(15);
			END IF;	
--	when ((its_me_lasted.fb==2) & (A==DA9) & (r_w.pin=='0') & (d15.pin=='0')) then
--		interrupt_flag := '0';
			IF cpuaddress(13 downto 12)="01" AND cpudata_in(15)='0' THEN	--DA9000
				irq_flag <= '0';
			END IF;	
--	when ((its_me_lasted.fb==2) & (A==DAA) & (r_w.pin=='0')) then
--		jeejee := d15.pin;
			IF cpuaddress(13 downto 12)="10" THEN	--DAA000
				int2ena <= cpudata_in(15);
			END IF;	
		END IF;	
--Init and shift gayle ID
		IF selgayleID='1' THEN
			IF cpuhwr='1' OR cpulwr='1' THEN
				shiftgid <= "11010000";			--gayleID = $D0
			ELSIF cpurd='1'	THEN
				shiftgid(7 downto 1) <= shiftgid(6 downto 0);
			END IF;	
		END IF;	
	END IF;
end process; 

------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------
--A600 IDE
	intide <= irq_flag;
------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------
--gary
--rd<=(cpurd AND cpuok) OR(NOT dmawr AND dma);
--hwr<=(NOT (cpuuds OR cpurd)AND cpuok) OR (dmawr AND dma);
--lwr<=(NOT (cpulds OR cpurd)AND cpuok) OR (dmawr AND dma);
rd<=cpurd OR(NOT dmawr AND dma);
hwr<=cpuhwr OR (dmawr AND dma);
lwr<=cpulwr OR (dmawr AND dma);

	
process(dma,cpuaddress,ovl,boot,e)
begin

--synchronize e clock to CPU available cycles
	IF clk'event AND clk = '1' THEN
		IF dma='0' THEN
			ecpu <= e;
		ELSIF dma='1' AND e='1' THEN
			ecpu <= e;
		END IF;	
	END IF;
		
--bus master logic
	IF dma='1' THEN		--bus slot allocated to agnus
		cpuok<='0';
	ELSIF (selreg OR selchip)='1' AND dmapri='1' THEN		--cpu wait state, dma has priority in register and chipram area
		cpuok<='0';
	ELSIF (selciaa OR selciab)='1' AND ecpu='0'	THEN		--cpu wait state, slow access to CIA's
		cpuok<='0';
	ELSIF selide='1' AND iderdy='0' THEN	
		cpuok<='0';
	ELSE		--bus slot allocated to cpu
		cpuok<='1';
	END IF;

--chip register bank and slowram address decode
--chipram, fastram, kickstart and bootrom address decode
	selslow <= '0';
	selreg <= '0';
	selgayleID <= '0';
	selgayle <= '0';
	selchip<='0';
	selfast <= '0';
	selkick<='0';
	selide<='0';
--	selkickovl<='0';
	selboot<='0';
	selciaa <= '0';
	selciab <= '0';
	selautoconfig <= '0';

	IF cpuaddress(23 downto 14)="1101101000" THEN	--ide port at &DA0000-DA3FFF 
		selide <= '1';
	ELSIF cpuaddress(23 downto 16)="11011110" THEN	--gayleID registers at &DE0000 - $DEFFFF
		selgayleID <= '1';
	ELSIF cpuaddress(23 downto 14)="1101101010" THEN	--gayle port at &DA8000-DABFFF 
		selgayle <= '1';
	ELSIF cpuaddress(23 downto 20)="1101" THEN	--chip registers at &D00000 - $D9FFFF
		IF dma='0' THEN	 
			selreg<='1';
		END IF;
	END IF;
	IF cpuaddress(23 downto 19)="11101" THEN	--$E80000	
		selautoconfig <= '1';
	END IF;

	IF dma='1' THEN		--agnus always accesses chipram
		selchip<='1';
	ELSIF cpuaddress(23)='0' THEN
		IF (cpuaddress(22 downto 21)="00") THEN	--chipram area
			IF boot='1' AND (cpuaddress(20 downto 12)=0)THEN
				selboot<='1';			--chipram area in boot mode
			ELSIF ovl='1' AND boot='0' THEN
				selkick<='1';			--kickstart when rom overlay
--				selkickovl<='1';
			ELSE		
				selchip<='1';			--chipram when no rom overlay
			END IF;
		ELSE
			selfast<='1';				--fastram at $200000 - $7FFFFF
		END IF;
	ELSE	
		IF cpuaddress(22)='0' THEN
			IF cpuaddress(21)='0' THEN
				selfast<='1';				--fastram at $800000 - $9FFFFF
			ELSE	
				IF cpuaddress(12)='0' THEN	--cia a address decode
					selciaa <= '1';
				END IF;
				IF cpuaddress(13)='0' THEN	--cia b address decode
					selciab <= '1';
				END IF;
			END IF;
		ELSE
--			IF cpuaddress(21 downto 19)="101" THEN	--$E80000	
--				selautoconfig <= '1';
--			END IF;
			IF cpuaddress(21 downto 19)="111" THEN	--$F80000 kickstart	
				selkick<='1';
			END IF;
			IF cpuaddress(21 downto 20)="00" THEN	--slow ram at $C00000 - $CFFFFF	
				selslow<='1';
			END IF;
			
--			IF cpuaddress(21 downto 16)="011110" THEN	--gayleID registers at &DE0000 - $DEFFFF
--				selgayleID <= '1';
--			ELSIF cpuaddress(21 downto 14)="01101010" THEN	--gayle port at &DA8000-DABFFF 
--				selgayle <= '1';
--			ELSIF cpuaddress(21 downto 20)="01" THEN	--chip registers at &D00000 - $D9FFFF
--				IF cpuaddress(21 downto 15)/="0110100" THEN	--ide port at &DA0000-DA3FFF 
--					selreg<='1';
--				END IF;
--			END IF;
		END IF;
	END IF;

	
------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------
--sdram bridge
--switch address and control bus between agnus and cpu
	IF dma='1' THEN
		ramaddress <= "000"&address_agnus;
	ELSE	
		ramaddress <= cpuaddress;
	END IF;
	
--	ramaddress(23 downto 1) <= "111"&mxaddress;
	enable <= '0';
	IF selchip='1'  THEN			--agnus always accesses chipram
--		ramaddress(23 downto 21) <= "010";
		enable <= '1';
		IF extra_chip='0' THEN
			ramaddress(20 downto 19)<="00";
		END IF;	
	END IF;	
	IF selfast='1' OR(selslow AND extra_slow AND (NOT cpuaddress(19) OR extra_chip))='1' THEN	--cpu accesses slowram
--		ramaddress(23 downto 20) <= "0010";
--		ramaddress(19)<=cpuaddress(19);
		enable <= '1';
	END IF;	
	IF(selkick AND(boot OR rd))='1' THEN		--kickstart JB Boot can write Kickrom
--	IF(selkick AND rd)='1' THEN		--kickstart
		ramaddress(23 downto 19) <= "11111";
		enable <= '1';
	END IF;	
end process; 

	n_we <= '0' WHEN (enable AND(hwr OR lwr))='1' ELSE '1'; 
	n_oe <= '0' WHEN (enable AND rd)='1' ELSE '1';
	n_ub <= '0' WHEN (enable AND(rd OR hwr))='1' ELSE '1';
	n_lb <= '0' WHEN (enable AND(rd OR lwr))='1' ELSE '1';
-- dataout multiplexer
	dataout <= data WHEN (enable AND rd)='1' ELSE "0000000000000000";

	auto_enable <= '1' WHEN selautoconfig='1' AND cfgin='0' AND cfgout='1' ELSE '0';  --AND cfgin='0'
	cfgin <= '0';
	PROCESS (clk)
	BEGIN
		IF rising_edge(clk) THEN
			IF reset='1' THEN
				cfgout <= '1';
			ELSIF auto_enable='1' AND cpuaddress(6 downto 1)="100100" AND cpuhwr='1' THEN	--4A
				cfgout <= '0';
			END IF;
		END IF;	
	
		CASE cpuaddress(6 downto 1) IS
--			WHEN "000000" => nibble <= "0010";	--00 128KB
--			WHEN "000001" => nibble <= "1110";	--02
			WHEN "000000" => nibble <= NOT "1110";	--00 onib
			WHEN "000001" => nibble <= NOT "0000";	--02 8MB unib
			WHEN "000010" => nibble <= "0000";	--04
			WHEN "000011" => nibble <= "0000";	--06
			WHEN "000100" => nibble <= "0000";	--08 
			WHEN "000101" => nibble <= "0000";	--0A
			WHEN "000110" => nibble <= "0000";	--0C
			WHEN "000111" => nibble <= "0000";	--0E
			
			WHEN "001000" => nibble <= "0001";	--10 
			WHEN "001001" => nibble <= "0010";	--12
			WHEN "001010" => nibble <= "0001";	--14
			WHEN "001011" => nibble <= "0010";	--16
			
			WHEN "001100" => nibble <= "0000";	--18 
			WHEN "001101" => nibble <= "0000";	--1A
			WHEN "001110" => nibble <= "0000";	--1C
			WHEN "001111" => nibble <= "0000";	--1E
			WHEN "010000" => nibble <= "0000";	--20 
			WHEN "010001" => nibble <= "0000";	--22
			WHEN "010010" => nibble <= "0000";	--24
			WHEN "010011" => nibble <= "0001";	--26
			
--			WHEN "010100" => nibble <= "0001";	--28 rom addr
			WHEN "010100" => nibble <= "0000";	--28 
			WHEN "010101" => nibble <= "0000";	--2A
			WHEN "010110" => nibble <= "0000";	--2C
			WHEN "010111" => nibble <= "0000";	--2E
			
			WHEN "011000" => nibble <= "0000";	--30 
			WHEN "011001" => nibble <= "0000";	--32
			WHEN "011010" => nibble <= "0000";	--34
			WHEN "011011" => nibble <= "0000";	--36
			WHEN "011100" => nibble <= "0000";	--38 
			WHEN "011101" => nibble <= "0000";	--3A
			WHEN "011110" => nibble <= "0000";	--3C
			WHEN "011111" => nibble <= "0000";	--3E
			
			
--			WHEN "100000" => nibble <= NOT "0000";	--40 
--			WHEN "100001" => nibble <= NOT "0000";	--42
			WHEN "100010" => nibble <= "0000";	--44
			WHEN "100011" => nibble <= "0000";	--46
			WHEN "100100" => nibble <= "0000";	--48 
			WHEN "100101" => nibble <= "0000";	--4A
			WHEN "100110" => nibble <= "0000";	--4C
			WHEN "100111" => nibble <= "0000";	--4E
			WHEN OTHERS => nibble <= "0000";
		END CASE;	
	END PROCESS;
END;	