------------------------------------------------------------------------------
------------------------------------------------------------------------------
--                                                                          --
-- Copyright (c) 2007-2009 Tobias Gubener <tobiflex@opencores.org>               -- 
--                                                                          --
-- This source file is free software: you can redistribute it and/or modify --
-- it under the terms of the GNU General Public License as published        --
-- by the Free Software Foundation, either version 3 of the License, or     --
-- (at your option) any later version.                                      --
--                                                                          --
-- This source file is distributed in the hope that it will be useful,      --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of           --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            --
-- GNU General Public License for more details.                             --
--                                                                          --
-- You should have received a copy of the GNU General Public License        --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.    --
--                                                                          --
------------------------------------------------------------------------------
------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity sdram is
port
	(
	ring		: out std_logic_vector(3 downto 0);
	sdata		: inout std_logic_vector(15 downto 0);
	sdaddr		: out std_logic_vector(11 downto 0);
	dqm			: out std_logic_vector(1 downto 0);
	sd_cs		: out std_logic_vector(3 downto 0);
	ba			: buffer std_logic_vector(1 downto 0);
	sd_we		: out std_logic;
	sd_ras		: out std_logic;
	sd_cas		: out std_logic;

	sysclk		: in std_logic;
	reset		: in std_logic;
	
	zdatawr		: in std_logic_vector(15 downto 0);
	zAddr		: in std_logic_vector(23 downto 0);
	zstate		: in std_logic_vector(2 downto 0);
	datawr		: in std_logic_vector(15 downto 0);
	rAddr		: in std_logic_vector(23 downto 0);
	rwr			: in std_logic;
	dwrL		: in std_logic;
	dwrU		: in std_logic;
	ZwrL		: in std_logic;
	ZwrU		: in std_logic;
	dma			: in std_logic;
	cpu_dma		: in std_logic;
	
	dataout		: out std_logic_vector(15 downto 0);
	zdataout	: out std_logic_vector(15 downto 0);
	c_14m		: out std_logic;
	zena_o		: out std_logic;
	c_28m		: out std_logic;
	c_7m		: out std_logic;
	reset_out	: out std_logic;
	IOB_clk		: out std_logic
	);
end;

architecture rtl of sdram is


signal initstate	:std_logic_vector(3 downto 0);
signal slow			:std_logic_vector(7 downto 0);
signal datard		:std_logic;
signal sdcom_ena	:std_logic;
signal isd_ras		:std_logic;
signal isd_cas		:std_logic;
signal isd_we 		:std_logic;
signal wr_int		:std_logic;
signal rd_int		:std_logic;
signal init_done	:std_logic;
signal Addr			:std_logic_vector(23 downto 0);
signal zmAddr		:std_logic_vector(23 downto 0);
signal isdaddr		:std_logic_vector(12 downto 0);
signal isd_cs		:std_logic_vector(3 downto 0);
signal datain		:std_logic_vector(15 downto 0);
signal casaddr		:std_logic_vector(10 downto 0);
signal casba		:std_logic_vector(1 downto 0);
signal wr 			:std_logic;
signal sdwrite 		:std_logic;
signal nextpix		:std_logic_vector(15 downto 0);


signal zena			:std_logic;
signal zcache		:std_logic_vector(63 downto 0);
signal zcache_addr	:std_logic_vector(23 downto 0);
signal zcache_recall	:std_logic;
signal fill_zcache	:std_logic;
signal match_zcache	:std_logic;
signal zvalid		:std_logic;
signal zequal		:std_logic;

signal rvalid		:std_logic;
signal rdataout		:std_logic_vector(15 downto 0);

signal V_cycle		:std_logic;
signal T_cycle		:std_logic;
signal R_cycle		:std_logic;
signal Z_cycle		:std_logic;

	type sdram_states is (ph0,ph1,ph2,ph3,ph4,ph5,ph6,ph7,ph8,ph9,ph10,ph11,ph12,ph13,ph14,ph15);
	signal sdram_state		: sdram_states;
	type pass_states is (nop,ras,cas);
	signal pass		: pass_states;

begin
	zmAddr <= NOT zAddr(23) & zAddr(22) & NOT zAddr(21) & zAddr(20 downto 0);
	Addr <= rAddr when R_cycle='1' else zmAddr;
	wr <= '0' when (R_cycle='1' AND rwr='0') OR (Z_cycle='1' AND zstate="011") else '1';
--	c_7m <= T_cycle;

	sd_ras <= isd_ras or sdcom_ena;
	sd_cas <= isd_cas or sdcom_ena;
	sd_we  <= isd_we  or sdcom_ena;
	sdaddr <= isdaddr(11 downto 0);
	sd_cs <= isd_cs when sdcom_ena='0' else "1111";
	
	zena_o <= '1' when zena='1' or match_zcache='1' else '0'; 
	
--	reset_out <= init_done AND reset;
	reset_out <= init_done;
	
--	process (sysclk, zmAddr, zcache_addr, zcache, zstate, zequal, zvalid, match_zcache, nextpix) begin
	process (sysclk, zmAddr, zcache_addr, zcache, zequal, zvalid, match_zcache, nextpix) begin
--		if zmaddr(22 downto 3)=zcache_addr(22 downto 3) then
--			zequal <='1';
--		else	
			zequal <='0';
--		end if;	
		if (zstate(2)='1' or zstate(1 downto 0)="01" or (zstate="00" and zequal='1' and zvalid='1')) then
			match_zcache <='1';
--			match_zcache <='0';
		else	
			match_zcache <='0';
		end if;	
--		case match_zcache&(zmaddr(2 downto 1)-zcache_addr(2 downto 1)) is
--			when "100"=>
				zdataout <= zcache(63 downto 48);
--			when "101"=>
--				zdataout <= zcache(47 downto 32);
--			when "110"=>
--				zdataout <= zcache(31 downto 16);
--			when "111"=>
--				zdataout <= zcache(15 downto 0);
--			when others=>
--				zdataout <= nextpix(15 downto 0);
--		end case;	
	end process;		
		
	process (rvalid, rdataout, nextpix)
    begin
--			if rvalid='1' then
				dataout <= rdataout;
--			else	
--				dataout <= nextpix;
--			end if;
	end process;		
	
	process (sysclk) begin
			if (sysclk'event and sysclk='1') then
				if sdram_state=ph0 then
					R_cycle <= '0';
					V_cycle <= '0';
					T_cycle <= '1';
					Z_cycle <= '0';
					
					IF slow(2 downto 0)=5 THEN
						slow <= slow+3;
					ELSE
						slow <= slow+1;
					END IF;
					IF dma='1' OR cpu_dma='1' THEN
						R_cycle <= '1';
						T_cycle <= '0';
					ELSIF slow(3 downto 0)="0001" THEN
						V_cycle <= '1';		--refresh cycle
					ELSE	
						Z_cycle <= '1';
					END IF;
--					case slow(3 downto 0) is
--						when "0001" => V_cycle <= '1';		--refresh cycle
--						when "0000"|"1000"|"0010"|"0100"|"1010"|"1100" => R_cycle <= '1';
--																		  T_cycle <= '0';
--						when others =>  Z_cycle <= '1';
--					end case;
				end if;
			end if;
	end process;		
	
	
--Datenbernahme
	process (sysclk, reset) begin
		if reset = '0' then
			zcache_recall <= '0';
			zvalid <= '0';
			zena <= '0';
	
		elsif (sysclk'event and sysclk='1') then
				nextpix <= sdata;
--				if zequal='1' and zstate="11" then
					zvalid <= '0';
--				end if;
				ring <= "0000";
					case sdram_state is	
						when ph0 =>	
										IOB_clk <= '0';
										ring(0) <='1';
						when ph1 =>	
										IOB_clk <= '1';
										c_28m <= '0';
										ring(1) <='1';
						when ph2 =>	
										IOB_clk <= '0';
--										zcache_recall <= match_zcache;
										zcache_recall <= zstate(2);
										ring(1) <='1';
						when ph3 =>	
										IOB_clk <= '1';
										c_28m <= '1';
										c_14m <= '0';

--										rvalid <= '0';
										ring(2) <='1';
						when ph4 =>	
										IOB_clk <= '0';
										ring(2) <='1';
						when ph5 =>
										IOB_clk <= '1';
										c_28m <= '0';
										ring(3) <='1';
						when ph6 =>	
										IOB_clk <= '0';
										ring(3) <='1';
						when ph7 =>	
										IOB_clk <= '1';
										c_28m <= '1';
										zena <= '0';
										c_14m <= '1'; 
										c_7m <= '0';
										ring(0) <='1';
										if Z_cycle='1'  then
											if zcache_recall='0' then
												zena <= '1';
----												if zstate="00" then
----													zcache_addr <= zmaddr;
														fill_zcache <= '1';
----													zvalid <= '0';
----												end if;
											end if;
										end if;
						when ph8 =>	
										IOB_clk <= '0';
										ring(0) <='1';
						when ph9 =>	
										IOB_clk <= '1';
										c_28m <= '0';
										if R_cycle='1' then
											rdataout <= nextpix;
											rvalid <= '1';
										end if;
										if fill_zcache='1' then
											zcache(63 downto 48) <= nextpix;
											zvalid <= '1';
										end if;
										ring(1) <='1';
						when ph10 =>	
										IOB_clk <= '0';
										if fill_zcache='1' then
											zcache(47 downto 32) <= nextpix;
										end if;
										fill_zcache <= '0';
										ring(1) <='1';
						when ph11 =>	
										IOB_clk <= '1';
										c_28m <= '1';
										c_14m <= '0';
										if fill_zcache='1' then
											zcache(31 downto 16) <= nextpix;
										end if;
										ring(2) <='1';
						when ph12 =>	
										IOB_clk <= '0';
										ring(2) <='1';
										if fill_zcache='1' then
											zcache(15 downto 0) <= nextpix;
										end if;
						when ph13 =>
										IOB_clk <= '1';
										c_28m <= '0';
										ring(3) <='1';
						when ph14 =>	
										IOB_clk <= '0';
										ring(3) <='1';
						when ph15 =>	
										IOB_clk <= '1';
										c_28m <= '1';
										zena <= '0';
										c_14m <= '1'; 
										c_7m <= '1';
										ring(0) <='1';
						when others =>	null;
					end case;	
			end if;
	end process;		
	
	

	process (sysclk, reset, sdwrite, datain) begin
		IF sdwrite='1' THEN
			sdata <= datain;
		ELSE
			sdata <= "ZZZZZZZZZZZZZZZZ";
		END IF;
		if reset = '0' then
			initstate <= (others => '0');
			init_done <= '0';
		else
			if (sysclk'event and sysclk='1') then
				sdcom_ena <= '1';
				sdwrite <= '0';
				case sdram_state is	--LATENCY=3
					when ph0 =>	pass <= ras;
								sdram_state <= ph1;
					when ph1 =>	sdcom_ena <= '0';
								sdram_state <= ph2;
					when ph2 =>	sdram_state <= ph3;
								casaddr <= addr(10 downto 0);
								casba <= ba;
								IF R_cycle='1' THEN
									datain <= datawr;
								ELSE	
									datain <= zdatawr;
								END IF;	
					when ph3 =>	pass <= cas;
								sdwrite <= '1';
								if V_cycle='1' then
									wr_int <= '1';
									rd_int <= '1';
								else
									wr_int <= wr ;
									rd_int <= '0';
								end if;
								sdram_state <= ph4;
					when ph4 =>	sdcom_ena <= '0';
								sdwrite <= '1';
								sdram_state <= ph5;
					when ph5 =>   
								sdram_state <= ph6;
					when ph6 =>	
								pass <= nop;
								sdram_state <= ph7;
					when ph7 =>	sdram_state <= ph8;
					when ph8 =>	sdram_state <= ph9;
					when ph9 =>	sdram_state <= ph10;
					when ph10 => sdram_state <= ph11;
					when ph11 => sdram_state <= ph12;
					when ph12 => sdram_state <= ph13;
					when ph13 => sdram_state <= ph14;
					when ph14 => sdram_state <= ph15;
					when ph15 =>	if initstate /= "1111" then
										initstate <= initstate+1;
									else
										init_done <='1';	
									end if;
								sdram_state <= ph0;
					when others =>
								sdram_state <= ph0;
				end case;	
			end if;
		end if;	
	end process;		


	
	process (initstate, pass, addr, wr, datain, init_done, casaddr, rd_int, wr_int, dwrU, dwrL, casba, Z_cycle, V_cycle, zmaddr) begin
		isdaddr<="XXXXXXXXXXXXX";
		isd_cs <= "1111"; --NOP
		isd_ras <= '1';
		isd_cas <= '1';
		isd_we <= '1';
		ba <= Addr(23 downto 22);

		if wr_int='0' then
			if Z_cycle='1' THEN
				dqm <= ZwrU& ZwrL;
			else
				dqm <= dwrU& dwrL;
			end if;
		else
			dqm <= "00";
		end if;	



		if init_done='0' then
			if pass=ras then
				case initstate is
					when "0010" => --PRECHARGE
						isdaddr(10) <= '1'; 	--all banks
						isd_cs <="0000";
						isd_ras <= '0';
						isd_cas <= '1';
						isd_we <= '0';
					when "0011"|"0100"|"0101"|"0110"|"0111"|"1000"|"1001"|"1010"|"1011"|"1100" => --AUTOREFRESH
						isd_cs <="0000"; 
						isd_ras <= '0';
						isd_cas <= '0';
						isd_we <= '1';
					when "1101" => --LOAD MODE REGISTER
						isd_cs <="0000";
						isd_ras <= '0';
						isd_cas <= '0';
						isd_we <= '0';
						ba <= "00";
--						isdaddr <= "0001000100010"; --BURST=4 LATENCY=2
						isdaddr <= "0001000110010"; --BURST=4 LATENCY=3
					when others =>	null;	--NOP
				end case;
			end if;
		else		
			case pass is
				when ras => --ACTIVE
					isdaddr <= "0"& addr(21 downto 10);
					isd_ras <= '0';
					isd_we <= '1';
					if V_cycle='1' then
						isd_cs <="0000"; --AUTOREFRESH
						isd_cas <= '0';
					else
						isd_cas <= '1';
						isd_cs <= "1110"; --ACTIVE
					end if;	
				when cas => --READ or Write
					ba <= casba;
					isdaddr <= '0'& '0' & '1' & "0" & casaddr(9 downto 1);--auto precharge
					isd_cs <= "1110"; 
					isd_ras <= '1';
					isd_cas <= rd_int;
					isd_we  <= wr_int;
				when others =>	null;	--NOP
			end case;
		end if;	
	end process;		
end;
